@file:JvmName("TransactionMethodsKt")
@file:OptIn(ExperimentalContracts::class, LowLevelApi::class)
@file:Suppress("FunctionName")

package ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp

import kotlin.contracts.ExperimentalContracts
import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlin.jvm.JvmName
import ru.landgrafhomyak.db.serdha0.api.LowLevelApi
import ru.landgrafhomyak.db.serdha0.api.queries._Query
import ru.landgrafhomyak.db.serdha0.api.runtime.Executor
import ru.landgrafhomyak.db.serdha0.api.runtime.InputRow
import ru.landgrafhomyak.db.serdha0.api.runtime.OutputRow
import ru.landgrafhomyak.db.serdha0.api.runtime.Transaction


@PublishedApi
internal inline fun <R> _safeAutoClose(onAbort: () -> Unit = {}, onSuccess: () -> Unit = {}, action: () -> R): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(onAbort, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onSuccess, InvocationKind.AT_MOST_ONCE)
	}
	val ret: R
	try {
		ret = action()
	} catch (e1: Throwable) {
		try {
			onAbort()
		} catch (e2: Throwable) {
			e1.addSuppressed(e2)
		}
		throw e1
	}
	onSuccess()
	return ret
}


suspend inline fun <R> Executor.transaction(@Suppress("REDUNDANT_INLINE_SUSPEND_FUNCTION_TYPE") scope: suspend (Transaction) -> R): R {
	val t = this._startTransaction()
	_safeAutoClose(onAbort = { t.rollback() }, onSuccess = { t.commit() }) {
		return scope(t)
	}
}

class ExpectedOneRowException() : Exception()
class TooManyRowsException() : Exception()

suspend inline fun <_qUE : Any, R> Transaction.selectOneOrError(
	compiledQuery: _Query.Params2Table<_qUE>,
	params: (qUE: _qUE, i: InputRow<_qUE>) -> Unit,
	result: (qUE: _qUE, o: OutputRow<_qUE>) -> R
): R {
	contract {
		callsInPlace(params, InvocationKind.EXACTLY_ONCE)
		callsInPlace(result, InvocationKind.EXACTLY_ONCE)
	}

	val i = this._executeQuery(compiledQuery)
	_safeAutoClose(onAbort = { i._abort() }) {
		params(compiledQuery.uExt, i)
	}
	val o = i._finish()
	val ret: R
	_safeAutoClose(onAbort = { o._abort() }, onSuccess = { o._finish() }) {
		if (!o._next())
			throw ExpectedOneRowException()
		ret = result(compiledQuery.uExt, o)
		if (o._next())
			throw TooManyRowsException()
	}
	return ret
}

suspend inline fun <_qUE : Any, R> Transaction.selectOneOrNull(
	compiledQuery: _Query.Params2Table<_qUE>,
	params: (qUE: _qUE, i: InputRow<_qUE>) -> Unit,
	result: (qUE: _qUE, o: OutputRow<_qUE>) -> R
): R? {
	contract {
		callsInPlace(params, InvocationKind.EXACTLY_ONCE)
		callsInPlace(result, InvocationKind.EXACTLY_ONCE)
	}

	val i = this._executeQuery(compiledQuery)
	_safeAutoClose(onAbort = { i._abort() }) {
		params(compiledQuery.uExt, i)
	}
	val o = i._finish()
	val ret: R
	_safeAutoClose(onAbort = { o._abort() }, onSuccess = { o._finish() }) {
		if (!o._next())
			return null
		ret = result(compiledQuery.uExt, o)
		if (o._next())
			throw TooManyRowsException()
	}
	return ret
}

suspend inline fun <_qUE : Any, E> Transaction.insertMany(
	compiledQuery: _Query.Table2Void<_qUE>,
	collection: Iterable<E>,
	transform: (qUE: _qUE, i: InputRow<_qUE>, e: E) -> Unit
) {
	contract {
		callsInPlace(transform, InvocationKind.UNKNOWN)
	}

	val i = this._executeQuery(compiledQuery)
	_safeAutoClose(onAbort = { i._abort() }) {
		for (e in collection) {
			i._next()
			transform(compiledQuery.uExt, i, e)
		}
	}
	i._finish()
}


suspend inline fun <_qUE : Any> Transaction.executeNoReturn(
	compiledQuery: _Query.Params2Table<_qUE>,
	params: (qUE: _qUE, i: InputRow<_qUE>) -> Unit,
) {
	contract {
		callsInPlace(params, InvocationKind.EXACTLY_ONCE)
	}

	val i = this._executeQuery(compiledQuery)
	_safeAutoClose(onAbort = { i._abort() }) {
		params(compiledQuery.uExt, i)
	}
	val o = i._finish()
	_safeAutoClose(onAbort = { o._abort() }, onSuccess = { o._finish() }) {
		if (o._next())
			throw TooManyRowsException()
	}
}

suspend inline fun <_qUE : Any> Transaction.executeNoReturn(
	compiledQuery: _Query.Params2Void<_qUE>,
	params: (qUE: _qUE, i: InputRow<_qUE>) -> Unit,
) {
	contract {
		callsInPlace(params, InvocationKind.EXACTLY_ONCE)
	}

	val i = this._executeQuery(compiledQuery)
	_safeAutoClose(onAbort = { i._abort() }) {
		params(compiledQuery.uExt, i)
	}
	i._finish()
}
